import json
import random
import os
import numpy as np
import itertools

from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.model_selection import train_test_split
from sklearn.model_selection import GridSearchCV
from sklearn.svm import SVR
from sklearn.metrics import mean_absolute_error
from sklearn.preprocessing import normalize, MinMaxScaler

import nltk
from   nltk.corpus import sentiwordnet as swn
from   nltk.corpus import stopwords
from   matplotlib import pyplot as plt


class ImportanceEstimationModel(object):


    def load_data(self, data_path):
        data = json.load(open(data_path))

        train_data = []
        test_data  = [] 
        ##### YOUR CODE HERE ######
        data = json.load(open(data_path))

        debates = list(set(map(lambda x: x['_debate_id'], data)))
        train_debates, test_debates = train_test_split(debates, test_size=0.2)

        train_instances = list(filter(lambda x: x['_debate_id'] in train_debates, data))
        test_instances  = list(filter(lambda x: x['_debate_id'] in test_debates, data))
        

        train_data = [(list(x['_argument_sentences'].values()), x['_claim']) for x in train_instances]
        test_data  = [(list(x['_argument_sentences'].values()), x['_claim']) for x in test_instances]

        ###########################
        return train_data, test_data

    def print_train_sample(self, train_data):
        sample = random.choice(train_data)
        premises = sample[0]
        conc   = sample[1]
        print('Conclusion:', conc)
        print('Premises:')
        for idx, premise in enumerate(premises[0:5]):
            print(idx+1, '.', premise)


    def _build_tfidf_model(self, texts):
        tfidf = TfidfVectorizer()
        tfidf_model = tfidf.fit(texts)
        return tfidf_model

    def _sentiment_features(self, premise):
        premise_words = nltk.word_tokenize(premise)

        num_of_positive_words = 0
        num_of_negative_words = 0
        num_of_neutral_words  = 0
        for word in premise_words:
            synsets = list(swn.senti_synsets(word))
            if len(synsets) == 0:
                num_of_neutral_words +=1
            else:
                syn = synsets[0]
                if syn.pos_score() > syn.neg_score():
                    num_of_positive_words +=1
                elif syn.pos_score() < syn.neg_score():
                    num_of_negative_words +=1
                else:
                    num_of_neutral_words+=1
        
        return num_of_positive_words, num_of_negative_words , num_of_neutral_words


    def _num_of_words_feature(self, premise):
        number_of_premise_words = 0
        ##### YOUR CODE HERE #######
        premise_words = nltk.word_tokenize(premise)
        return len(premise_words)

    def _tfidf_features(self, premise):
        avg_tfidf_feature = 0
        max_tfidf_feature = 0
        ##### YOUR CODE HERE #######
        premise_words = nltk.word_tokenize(premise)
        
        # Avg. tfidf
        tfidf_vector = self.tfidf_model.transform([premise])
        avg_tfidf_feature = np.sum(tfidf_vector.toarray())/len(premise_words)
        max_tfidf_feature = np.max(tfidf_vector.toarray())
        return avg_tfidf_feature, max_tfidf_feature

    def _premise_features(self, premise, premises_text):
        # Number of words
        num_of_words_feature = self._num_of_words_feature(premise)
        
        # Avg. Max. tfidf
        avg_tfidf_feature, max_tfidf_feature = self._tfidf_features(premise)
        
        # Number of postive/negative/neutral words
        num_of_positive_words, num_of_negative_words , num_of_neutral_words  = self._sentiment_features(premise)
        
        return [num_of_words_feature, 
                avg_tfidf_feature, 
                max_tfidf_feature,
                num_of_positive_words, 
                num_of_negative_words, 
                num_of_neutral_words]

    def _instance_features(self, premises):
        premises_text = ' '.join(premises)
        premises_features = np.array([self._premise_features(premise, premises_text) for premise in premises])
        return premises_features

    def instance_scores(self, premises, conclusion):
        premises_scores = []
        ##### YOUR CODE HERE ########

        #Download stopwords list
        nltk.download('stopwords')
        stop_words = set(stopwords.words('english'))

        for premise in premises:
            premise_tokens   = set(nltk.word_tokenize(premise))
            conclusion_tokens = set(nltk.word_tokenize(conclusion))

            premise_tokens = set(filter(lambda x: x not in stop_words, premise_tokens))
            conclusion_tokens = set(filter(lambda x: x not in stop_words, conclusion_tokens))

            shared_tokens  = premise_tokens.intersection(conclusion_tokens)
            premises_scores.append(len(shared_tokens))
        
        ############################
        return premises_scores

    def feature_representation(self, data, mask=None):
        # 1. build a tf-idf model over the training data
        arguments = [' '.join(argument[0]) for argument in data]
        self.tfidf_model = self._build_tfidf_model(arguments)

        # 2. Encode training data into features
        self.train_X = []
        self.train_Y = []

        for argument in data:
            premises     = argument[0]
            conclusion = argument[1]

            premises_vectors = self._instance_features(premises)
            premises_labels  = self.instance_scores(premises, conclusion)

            for premise_vector, premise_label in zip(premises_vectors, premises_labels):
                self.train_X.append(premise_vector)
                self.train_Y.append(premise_label)


        self.train_X = np.array(self.train_X)
        self.train_Y = np.array(self.train_Y)

        return self.train_X, self.train_Y

    def train_grid_search_svr(self, train_X, train_Y):
        svr_params = {'C': [0.001, 0.1, 1.0, 10, 100]}
        best_svr = None
        best_score = None
        ##### YOUR CODE HERE ########
        svr = SVR()
        
        clf = GridSearchCV(svr, svr_params, cv=5, scoring='neg_mean_absolute_error', return_train_score=False)
        clf.fit(train_X, train_Y)

        best_svr = clf.best_estimator_
        
        self.best_svr = best_svr

        best_score =  clf.best_score_

        return best_svr, best_score


    def features_evaluation(self, train_data, test_data):
        train_X, train_Y = self.feature_representation(train_data)
        mrr_values = {}

        #Use only tf-idf features
        mask = [0,1,1,0,0,0]
        train_X1 = [list(itertools.compress(x, mask)) for x in train_X]
        self.train_svr(train_X1, train_Y)
        mrr_values['tf-idf'] = self.mrr_evaluation(test_data, mask)

        mask = [0,0,0,1,1,1]
        train_X2 = [list(itertools.compress(x, mask)) for x in train_X]
        self.train_svr(train_X2, train_Y)
        mrr_values['sentiment'] = self.mrr_evaluation(test_data, mask)

        mask = [1,0,0,0,0,0]
        train_X3 = [list(itertools.compress(x, mask)) for x in train_X]
        self.train_svr(train_X3, train_Y)
        mrr_values['number of words'] = self.mrr_evaluation(test_data, mask)

        mask = [1,1,1,1,1,1]
        train_X4 = [list(itertools.compress(x, mask)) for x in train_X]
        self.train_svr(train_X4, train_Y)
        mrr_values['all'] = self.mrr_evaluation(test_data, mask)


        return mrr_values


    def mrr_evaluation(self, test_data, mask=None):
        mrr_value = 0
        for sample in test_data:
            premises     = sample[0]
            conclusion = sample[1]

            #Predict scores of each premise
            premises_vectors = self._instance_features(premises)
            if mask:
                premises_vectors = [list(itertools.compress(x, mask)) for x in premises_vectors]

            premises_labels  = self.instance_scores(premises, conclusion)
            premises_scores  = self.best_svr.predict(premises_vectors)

            #Sort premises based on the score
            scores_labels_list = list(zip(premises_scores, premises_labels))
            sorted_premises = sorted(scores_labels_list, key= lambda x : -x[0])

            rank_of_first_item = 0

            for item in sorted_premises:
                rank_of_first_item += 1
                if item[1] > 0:
                    break

            mrr_value += 1/rank_of_first_item


        return mrr_value/len(test_data)